package com.geekcap.informit.neo4jsample;

import org.neo4j.graphdb.*;
import org.neo4j.graphdb.factory.GraphDatabaseFactory;

import java.util.Map;

/**
 * Demonstrates how to use Cypher to query the user/movie database we've been working with in this example
 */
public class Neo4jCypherSample
{
    /**
     * Node labels: a node can be a USER or it can be a MOVIE
     */
    public enum Labels implements Label
    {
        USER,
        MOVIE;
    }

    /**
     * Relationships: USERs can have an IS_FRIEND_OF relationship with other USERs and USERs can have a
     * HAS_SEEN relationship with MOVIEs
     */
    public enum RelationshipTypes implements RelationshipType
    {
        IS_FRIEND_OF,
        HAS_SEEN;
    }

    public static Relationship seeMovie( Node user, Node movie, int stars )
    {
        Relationship relationship = user.createRelationshipTo( movie, RelationshipTypes.HAS_SEEN );
        relationship.setProperty( "stars", stars );
        return relationship;
    }

    public static void main( String[] args )
    {
        // Create an embedded database in the "data" directory (relative to where this class is being run)
        GraphDatabaseService graphDB = new GraphDatabaseFactory().newEmbeddedDatabase("data");

        // Setup our data
        try( Transaction txn = graphDB.beginTx() )
        {
            // Create four users: Steve, Linda, Michael, and Rebecca
            Node steve = graphDB.createNode(Labels.USER);
            steve.setProperty("name", "Steve");
            Node linda = graphDB.createNode(Labels.USER);
            linda.setProperty("name", "Linda");
            Node michael = graphDB.createNode(Labels.USER);
            michael.setProperty("name", "Michael");
            Node rebecca = graphDB.createNode(Labels.USER);
            rebecca.setProperty("name", "Rebecca");

            // Create relationships between them
            steve.createRelationshipTo(michael, RelationshipTypes.IS_FRIEND_OF);
            steve.createRelationshipTo(rebecca, RelationshipTypes.IS_FRIEND_OF);
            steve.createRelationshipTo(linda, RelationshipTypes.IS_FRIEND_OF);
            michael.createRelationshipTo(rebecca, RelationshipTypes.IS_FRIEND_OF);
            michael.createRelationshipTo(linda, RelationshipTypes.IS_FRIEND_OF);

            // Create three movies: Divergent, Big Hero 6, and Cinderella
            Node divergent = graphDB.createNode(Labels.MOVIE);
            divergent.setProperty("name", "Divergent");
            Node hero = graphDB.createNode(Labels.MOVIE);
            hero.setProperty("name", "Big Hero 6");
            Node cinderella = graphDB.createNode(Labels.MOVIE);
            cinderella.setProperty("name", "Cinderella");
            Node interview = graphDB.createNode(Labels.MOVIE);
            interview.setProperty("name", "The Interview");

            // Create relationships between users and movies
            seeMovie(steve, divergent, 5);
            seeMovie(steve, hero, 5);
            seeMovie(steve, cinderella, 4);
            seeMovie(steve, interview, 1);
            seeMovie(rebecca, hero, 5);
            seeMovie(rebecca, cinderella, 5);
            seeMovie(michael, hero, 5);
            seeMovie(michael, cinderella, 3);
            seeMovie(linda, divergent, 4);
            seeMovie(linda, hero, 5);
            seeMovie(linda, cinderella, 5);

            // Success!
            txn.success();
        }

        //
        // Cypher queries
        //

        // Find all users
        System.out.println( "Users" );
        try( Transaction txn = graphDB.beginTx();
             Result results = graphDB.execute( "match( user :USER ) return user.name" ) )
        {
            while( results.hasNext() )
            {

                Map<String,Object> result = results.next();
                System.out.println( "\t" + result.get( "user.name" ) );
            }
            txn.success();
        }

        // Find all movies
        System.out.println( "Movies" );
        try( Transaction txn = graphDB.beginTx();
             Result results = graphDB.execute( "match( movie :MOVIE ) return movie.name" ) )
        {
            while( results.hasNext() )
            {

                Map<String,Object> result = results.next();
                System.out.println( "\t" + result.get( "movie.name" ) );
            }
            txn.success();
        }

        // Find Michael's friends
        System.out.println( "Michael's friends" );
        try( Transaction txn = graphDB.beginTx();
             Result results = graphDB.execute( "match( michael:USER { name: 'Michael' } )-[:IS_FRIEND_OF]-(friend) return friend.name" ) )
        {
            while( results.hasNext() )
            {

                Map<String,Object> result = results.next();
                System.out.println( "\t" + result.get( "friend.name" ) );
            }
            txn.success();
        }

        // Find Michael's movies
        System.out.println( "Michael's movies" );
        try( Transaction txn = graphDB.beginTx();
             Result results = graphDB.execute( "match( michael:USER { name: 'Michael' } )-[:HAS_SEEN]-(movie) return movie.name" ) )
        {
            while( results.hasNext() )
            {

                Map<String,Object> result = results.next();
                System.out.println( "\t" + result.get( "movie.name" ) );
            }
            txn.success();
        }

        // Find Michael's friend's movies
        System.out.println( "Michael's friend's movies" );
        try( Transaction txn = graphDB.beginTx();
             Result results = graphDB.execute( "match( michael:USER { name: 'Michael' } )-[:IS_FRIEND_OF]-(friend)-[:HAS_SEEN]->(movie) return movie.name" ) )
        //   Result results = graphDB.execute( "match( :USER { name: 'Michael' } )-[:IS_FRIEND_OF]-()-[:HAS_SEEN]->(movie) return movie.name" ) )
        {
            while( results.hasNext() )
            {

                Map<String,Object> result = results.next();
                System.out.println( "\t" + result.get( "movie.name" ) );
            }
            txn.success();
        }

        // Find Michael's friend's movies (distinct)
        System.out.println( "Michael's friend's movies (distinct)" );
        try( Transaction txn = graphDB.beginTx();
             Result results = graphDB.execute( "match( michael:USER { name: 'Michael' } )-[:IS_FRIEND_OF]-(friend)-[:HAS_SEEN]->(movie) return DISTINCT movie.name" ) )
        {
            while( results.hasNext() )
            {

                Map<String,Object> result = results.next();
                System.out.println( "\t" + result.get( "movie.name" ) );
            }
            txn.success();
        }

        // Find Michael's friend's movies (sorted)
        System.out.println( "Michael's friend's movies (sorted by name)" );
        try( Transaction txn = graphDB.beginTx();
             Result results = graphDB.execute( "match( michael:USER { name: 'Michael' } )-[:IS_FRIEND_OF]-(friend)-[:HAS_SEEN]->(movie) return DISTINCT movie.name ORDER BY movie.name" ) )
        {
            while( results.hasNext() )
            {

                Map<String,Object> result = results.next();
                System.out.println( "\t" + result.get( "movie.name" ) );
            }
            txn.success();
        }

        // Find Michael's friend's movies (sorted descending)
        System.out.println( "Michael's friend's movies (sorted by name descending)" );
        try( Transaction txn = graphDB.beginTx();
             Result results = graphDB.execute( "match( michael:USER { name: 'Michael' } )-[:IS_FRIEND_OF]-(friend)-[:HAS_SEEN]->(movie) return DISTINCT movie.name ORDER BY movie.name DESC" ) )
        {
            while( results.hasNext() )
            {

                Map<String,Object> result = results.next();
                System.out.println( "\t" + result.get( "movie.name" ) );
            }
            txn.success();
        }

        // Find Michael's friend's movies that Michael hasn't seen
        System.out.println( "Michael's friend's movies (that he hasn't seen)" );
        try( Transaction txn = graphDB.beginTx();
             Result results = graphDB.execute( "match( michael:USER { name: 'Michael' } )-[:IS_FRIEND_OF]-(friend)-[:HAS_SEEN]->(movie) WHERE NOT (michael)-[:HAS_SEEN]-(movie) return DISTINCT movie.name ORDER BY movie.name" ) )
        {
            while( results.hasNext() )
            {

                Map<String,Object> result = results.next();
                System.out.println( "\t" + result.get( "movie.name" ) );
            }
            txn.success();
        }
    }

}
