package com.geekcap.informit.neo4jsample;

import org.neo4j.graphdb.*;
import org.neo4j.graphdb.factory.GraphDatabaseFactory;

import java.util.Map;

/**
 * More advanced Cypher example
 */
public class Neo4jMovieCatalogSample
{
    /**
     * Node labels: a node can be a USER or it can be a MOVIE
     */
    public enum Labels implements Label
    {
        MOVIE,
        TAG,
        CHARACTER;
    }

    /**
     * Relationships: USERs can have an IS_FRIEND_OF relationship with other USERs and USERs can have a
     * HAS_SEEN relationship with MOVIEs
     */
    public enum RelationshipTypes implements RelationshipType
    {
        HAS_TAG,
        PERFORMS_IN;
    }

    public static void main( String[] args )
    {
        GraphDatabaseService graphDB = new GraphDatabaseFactory().newEmbeddedDatabase("data");

        // Setup a movie graph
        try( Transaction txn = graphDB.beginTx() )
        {
            // Create characters
            Node ironMan = graphDB.createNode( Labels.CHARACTER );
            ironMan.setProperty( "name", "Iron Man" );
            Node captainAmerica = graphDB.createNode( Labels.CHARACTER );
            captainAmerica.setProperty( "name", "Captain America" );
            Node hulk = graphDB.createNode( Labels.CHARACTER );
            hulk.setProperty( "name", "The Hulk" );
            Node thor = graphDB.createNode( Labels.CHARACTER );
            thor.setProperty( "name", "Thor" );
            Node wolverine = graphDB.createNode( Labels.CHARACTER );
            wolverine.setProperty( "name", "Wolverine" );

            // Create movie tags
            Node actionTag = graphDB.createNode( Labels.TAG );
            actionTag.setProperty( "name", "Action" );
            Node marvelTag = graphDB.createNode( Labels.TAG );
            marvelTag.setProperty( "name", "Marvel" );
            Node superHeroTag = graphDB.createNode( Labels.TAG );
            superHeroTag.setProperty( "name", "Super Hero" );
            Node xmenTag = graphDB.createNode( Labels.TAG );
            xmenTag.setProperty( "name", "X-Men" );
            Node newReleaseTag = graphDB.createNode( Labels.TAG );
            newReleaseTag.setProperty( "name", "New Release" );

            // Create movies
            Node avengers = graphDB.createNode( Labels.MOVIE );
            avengers.setProperty( "name", "The Avengers" );
            avengers.createRelationshipTo( marvelTag, RelationshipTypes.HAS_TAG );
            avengers.createRelationshipTo( actionTag, RelationshipTypes.HAS_TAG );
            avengers.createRelationshipTo( superHeroTag, RelationshipTypes.HAS_TAG );
            ironMan.createRelationshipTo( avengers, RelationshipTypes.PERFORMS_IN );
            captainAmerica.createRelationshipTo( avengers, RelationshipTypes.PERFORMS_IN );
            hulk.createRelationshipTo( avengers, RelationshipTypes.PERFORMS_IN );
            thor.createRelationshipTo( avengers, RelationshipTypes.PERFORMS_IN );

            Node avengers2 = graphDB.createNode( Labels.MOVIE );
            avengers2.setProperty( "name", "The Avengers: Age of Ultron" );
            avengers2.createRelationshipTo( marvelTag, RelationshipTypes.HAS_TAG );
            avengers2.createRelationshipTo( actionTag, RelationshipTypes.HAS_TAG );
            avengers2.createRelationshipTo( superHeroTag, RelationshipTypes.HAS_TAG );
            avengers2.createRelationshipTo( newReleaseTag, RelationshipTypes.HAS_TAG );
            ironMan.createRelationshipTo( avengers2, RelationshipTypes.PERFORMS_IN );
            captainAmerica.createRelationshipTo( avengers2, RelationshipTypes.PERFORMS_IN );
            hulk.createRelationshipTo( avengers2, RelationshipTypes.PERFORMS_IN );
            thor.createRelationshipTo( avengers2, RelationshipTypes.PERFORMS_IN );

            Node ironManMovie = graphDB.createNode( Labels.MOVIE );
            ironManMovie.setProperty( "name", "Iron Man" );
            ironManMovie.createRelationshipTo( marvelTag, RelationshipTypes.HAS_TAG );
            ironManMovie.createRelationshipTo( actionTag, RelationshipTypes.HAS_TAG );
            ironManMovie.createRelationshipTo( superHeroTag, RelationshipTypes.HAS_TAG );
            ironMan.createRelationshipTo( ironManMovie, RelationshipTypes.PERFORMS_IN );

            Node theWolverine = graphDB.createNode( Labels.MOVIE );
            theWolverine.setProperty( "name", "The Wolverine" );
            theWolverine.createRelationshipTo( xmenTag, RelationshipTypes.HAS_TAG );
            theWolverine.createRelationshipTo( actionTag, RelationshipTypes.HAS_TAG );
            theWolverine.createRelationshipTo( superHeroTag, RelationshipTypes.HAS_TAG );
            wolverine.createRelationshipTo( theWolverine, RelationshipTypes.PERFORMS_IN );

            Node home = graphDB.createNode( Labels.MOVIE );
            home.setProperty( "name", "Home" );
            home.createRelationshipTo( newReleaseTag, RelationshipTypes.HAS_TAG );

            // Success!
            txn.success();
        }

        //
        // Cypher Queries
        //

        // Find all movies
        System.out.println( "Movies" );
        try( Transaction txn = graphDB.beginTx();
             Result results = graphDB.execute( "match( movie :MOVIE ) return movie.name" ) )
        {
            while( results.hasNext() )
            {

                Map<String,Object> result = results.next();
                System.out.println( "\t" + result.get( "movie.name" ) );
            }
            txn.success();
        }

        // Find all Marvel movies
        System.out.println( "Marvel Movies" );
        try( Transaction txn = graphDB.beginTx();
             Result results = graphDB.execute( "MATCH(marvelMovies:TAG {name:'Marvel'})<-[:HAS_TAG]-(movie) RETURN movie.name" ) )
            //Result results = graphDB.execute( "MATCH(marvelMovies:TAG)-[:HAS_TAG]-(movie) WHERE marvelMovies.name ='Marvel' RETURN movie.name" ) )
        {
            while( results.hasNext() )
            {

                Map<String,Object> result = results.next();
                System.out.println( "\t" + result.get( "movie.name" ) );
            }
            txn.success();
        }

        // Find all New Release Marvel movies
        System.out.println( "New Release Marvel Movies" );
        try( Transaction txn = graphDB.beginTx();
             Result results = graphDB.execute( "MATCH(newRelease:TAG {name:'New Release'}),(marvelMovies:TAG {name:'Marvel'})-[:HAS_TAG]-(movie) WHERE (newRelease)-[:HAS_TAG]-(movie) RETURN movie.name" ) )
        {
            while( results.hasNext() )
            {

                Map<String,Object> result = results.next();
                System.out.println( "\t" + result.get( "movie.name" ) );
            }
            txn.success();
        }

        // Find all New Release Super Hero Marvel movies
        System.out.println( "New Release Marvel Super Hero Movies" );
        try( Transaction txn = graphDB.beginTx();
             Result results = graphDB.execute( "MATCH(superHeroes:TAG {name:'Super Hero'}),(newRelease:TAG {name:'New Release'}),(marvelMovies:TAG {name:'Marvel'})-[:HAS_TAG]-(movie) WHERE (newRelease)-[:HAS_TAG]-(movie) AND (superHeroes)-[:HAS_TAG]-(movie) RETURN movie.name" ) )
        {
            while( results.hasNext() )
            {

                Map<String,Object> result = results.next();
                System.out.println( "\t" + result.get( "movie.name" ) );
            }
            txn.success();
        }

        // Find all Marvel movies with Thor: all movies that Thor PERFORMS_IN AND all movies that HAS_TAG Marvel
        System.out.println( "Marvel Movies with Thor" );
        try( Transaction txn = graphDB.beginTx();
             Result results = graphDB.execute( "MATCH(marvelMovies:TAG {name:'Marvel'})-[:HAS_TAG]-(movie)-[:PERFORMS_IN]-(thor:CHARACTER {name:'Thor'}) RETURN movie.name" ) )
        {
            while( results.hasNext() )
            {

                Map<String,Object> result = results.next();
                System.out.println( "\t" + result.get( "movie.name" ) );
            }
            txn.success();
        }

        // Find all New Release Marvel movies with Thor
        System.out.println( "New Release Marvel Movies with Thor" );
        try( Transaction txn = graphDB.beginTx();
             Result results = graphDB.execute( "MATCH(marvelMovies:TAG {name:'Marvel'}),(newRelease:TAG {name:'New Release'})-[:HAS_TAG]-(movie)-[:PERFORMS_IN]-(thor:CHARACTER {name:'Thor'}) WHERE (marvelMovies)-[:HAS_TAG]-(movie) RETURN movie.name" ) )
        {
            while( results.hasNext() )
            {

                Map<String,Object> result = results.next();
                System.out.println( "\t" + result.get( "movie.name" ) );
            }
            txn.success();
        }

        // Find all New Release Marvel movies with Thor and the Hulk
        System.out.println( "New Release Marvel Movies with Thor and the Hulk" );
        try( Transaction txn = graphDB.beginTx();
             Result results = graphDB.execute( "MATCH(marvelMovies:TAG {name:'Marvel'}),(newRelease:TAG {name:'New Release'})-[:HAS_TAG]-(movie)-[:PERFORMS_IN]-(thor:CHARACTER {name:'Thor'}),(hulk:CHARACTER {name:'The Hulk'}) WHERE (marvelMovies)-[:HAS_TAG]-(movie) AND (movie)-[:PERFORMS_IN]-(hulk) RETURN movie.name" ) )
        {
            while( results.hasNext() )
            {

                Map<String,Object> result = results.next();
                System.out.println( "\t" + result.get( "movie.name" ) );
            }
            txn.success();
        }

        // Find all New Release Marvel movies with Thor and the Hulk
        System.out.println( "New Release Marvel Movies with Thor and the Hulk (alternative query)" );
        try( Transaction txn = graphDB.beginTx();
             Result results = graphDB.execute( "MATCH(thor:CHARACTER {name:'Thor'}),(hulk:CHARACTER {name:'The Hulk'}),(marvelMovies:TAG {name:'Marvel'}),(newRelease:TAG {name:'New Release'})-[:HAS_TAG]-(movie) WHERE (marvelMovies)-[:HAS_TAG]-(movie) AND (movie)-[:PERFORMS_IN]-(hulk) AND (movie)-[:PERFORMS_IN]-(thor) RETURN movie.name" ) )
        {
            while( results.hasNext() )
            {

                Map<String,Object> result = results.next();
                System.out.println( "\t" + result.get( "movie.name" ) );
            }
            txn.success();
        }


    }
}
